/******************************************************************************

 Autor: Martin Kot
 Studijni cislo: KOT119
 Predmet: Operacni systemy

 Projekt cislo 1 - implementace algoritmu producer/consumer pomoci semaforu
 
 Program po spusteni vytvori synovsky proces, ktery do sdilene oblasti pameti 
 umistuje znaky ze souboru. Jmeno zdrojoveho souboru se zadava jako prvni 
 argument programu. Otcovsky proces tyto znaky ze sdilene pameti cte a vypisuje
 je na standardni vystup.

******************************************************************************/
  
#include <sys/sem.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <ctype.h>

#define MEM_SIZE 100 /*Velikost sdilene pameti*/
#define SEM_COUNT 3  /*Pocet nutnych semaforu*/

/*Hodnota pro rozliseni 3 semaforu*/
#define FULL 0
#define EMPTY 1
#define MUTEX 2

/*Zvetsi hodnotu semaforu o 1*/
void up(int sid, int semnumber){
  int semval;
  struct sembuf sem_up={semnumber,1,SEM_UNDO};
  semop(sid, &sem_up, 1);
}

/*Zmensi hodnotu semaforu o 1. Pokud je hodnota 0, uspi proces dokud se hodnota
  nezvetsi.*/
void down(int sid, int semnumber){
  int semval;
  struct sembuf sem_down={semnumber,-1,SEM_UNDO};
  semop(sid, &sem_down, 1);
}


/*Funkce na ziskani hodnoty semaforu*/
int getSemVal(int semid, int semnumber){
  int semval;
  semval=semctl(semid,semnumber,GETVAL,0);
}  
  
/*Vytvori semafory a nastavi jejich pocatecni hodnoty*/
int makeSemafors(char *argv){
  int semafor;
  key_t key;

  /*Inicializace generatoru pseudonahodnych cisel*/
  srand((unsigned) time(NULL));       
  key = ftok(argv,(char) (rand()%255)); /*Nehodny klic pro semafor*/

  /*Vytvoreni semaforu*/
  if((semafor = semget(key, SEM_COUNT, IPC_CREAT|0666)) == -1){
    fprintf(stderr,"Semafory s danym klicem jiz existuji.\n");
    exit(0);
  }

  /*Nastaveni pocatecnich hodnot semaforu*/
  if(semctl(semafor,EMPTY,SETVAL,MEM_SIZE)==1) {
    perror("EMPTY semafor: ");
  }
  if (semctl(semafor,FULL,SETVAL,0)==-1){
    perror("FULL semafor: ");
  }
  if(semctl(semafor,MUTEX,SETVAL,1)==-1){
    perror("MUTEX semafor: ");
  }

  return semafor;
}

/*Vytvori sdilenou pamet*/
int makeMemory(char *argv){
  int pamet_id;
  key_t key;

  /*Inicializace generatoru pseudonahodnych cisel*/
  srand((unsigned) time(NULL)); 
  key = ftok(argv,(char) (rand()%255)); /*Nehodny klic pro semafor*/

  if((pamet_id = shmget(key,MEM_SIZE,IPC_CREAT|IPC_EXCL|0666))==-1){
    /*Pamet s danym klicem jiz existuje. Pokud je pristupna, bude ji program 
      uzivat*/
    if((pamet_id = shmget(key,MEM_SIZE,0))==-1) {
      /*Pamet neni pristupna*/
      fprintf(stderr,"Nelze vytvorit sdilenou pamet.\n");
      exit(1);
    }
  }
  return pamet_id;
}

/*Cte znaky ze souboru a umistuje je do sdilene pameti*/
void produce(int sem_id, char *file, char *shm_uk) {
  static int prvek = 0;
  FILE *zdroj;
  char c='c';

  if((zdroj = fopen(file,"r"))==NULL){
    /*Nepovedlo se otevrit soubor, umisti do sdilene pameti jen zarazku*/
    fprintf(stderr,"Nepodarilo se otevrit soubor \"%s\".\n",file);
    down(sem_id,EMPTY);
    down(sem_id,MUTEX);
    *(shm_uk+prvek)='\x0';
    up(sem_id,MUTEX);
    up(sem_id,FULL);
  }
  else {
    /*Prochazi soubor a umistuje znaky do sdilene pameti*/
    while(c!=EOF && !feof(zdroj) && c!='\x0'){
      fread(&c,1,sizeof(char),zdroj);
      if(feof(zdroj)||c==EOF){
        c='\x0';
      }
      down(sem_id,EMPTY);
      down(sem_id,MUTEX);
      *(shm_uk+prvek)=c;
      up(sem_id,MUTEX);
      up(sem_id,FULL);
      prvek = (++prvek<MEM_SIZE)?prvek:prvek%MEM_SIZE;
    }
    fclose(zdroj);
  }
  /*Pocka, az budou znaky ze sdilene pameti odebrany. Jinak by se vratily 
    hodnoty semaforu na predchozi hodnoty*/
  while(getSemVal(sem_id,FULL)!=0);
}

/*Cte znaky ze sdilene pameti dokud nanarazi na znak \x0 a vypisuje je na
  standardni vystup*/
void consume(int sem_id, char *shm_uk) {
  static int cteny = 0;
  char c;
  do {
    down(sem_id,FULL);
    down(sem_id,MUTEX);
    c = *(shm_uk+cteny);
    up(sem_id,MUTEX);
    up(sem_id,EMPTY);
    cteny = (++cteny<MEM_SIZE)?cteny:cteny%MEM_SIZE;
    printf("%c",c);
  }while(c!='\x0');
}

/*Vypsani strucne napovedy*/
void printUsage(void){
  printf("Program kopiruje obsah souboru na standardni vystup.\n");
  printf("Jmeno souboru se zadava jako prvni argument programu.\n");
}



int main(int argc, char **argv){
  char *pamet;
  int semafor;
  int pid,vysl,pamet_id;

  if(argc < 2){
    /*Nebylo zadano jmeno souboru*/
    printUsage();
    return 0;
  } 

  semafor = makeSemafors(argv[0]);
  pamet_id = makeMemory(argv[0]); 
  
  /*Namapuje sdilenou pamet*/
  if((pamet = shmat(pamet_id,0,0))== (char *) -1) {
    fprintf(stderr,"Nelze primapovat pamet. %d\n",pamet_id);
    perror("");
    exit(1);
  }

  /*Vytvoreni synovskeho procesu*/
  if((pid=fork())==-1){
    fprintf(stderr,"Nelze vytvorit synovsky proces.\n");
  } else if (pid == 0) {
    /*Synovsky proces slouzi jako producent*/
    produce(semafor, argv[1],pamet);
    shmdt(pamet);
  } else {
    /*Otcovsky proces slouzi jako konzument*/
    consume(semafor,pamet);
    waitpid(pid,&vysl,0);
    shmdt(pamet);
    /*Zrusi semafory a sdilenou pamet*/
    semctl(semafor,0,IPC_RMID,0);
    shmctl(pamet_id, IPC_RMID, 0);
  }
  return 0;
}
    

