/******************************************************************************

 Autor: Martin Kot
 Studijni cislo: KOT119
 Predmet: Operacni systemy

 Projekt cislo 2 - implementace algoritmu jedici filozofove pomoci front zprav
 
 Program simuluje stridave jezeni a mysleni 5ti filozofu sedicich v kruhu. 
 Nikdy nesmi jist soucasne 2 filozofove sedici vedle sebe. Informace o jejich
 cinnosti se vypisuji na standardni vystup. Ukonceni programu se provadi 
 poslanim signalu SIG_INT (stisk CTRL+C). Po jeho prijeti jeste filozofove, 
 kteri chteli jist nebo jiz ji, doji a vypisi hlasku o svem konci.

******************************************************************************/
  
#include <sys/ipc.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <signal.h>
#include <sys/msg.h>

#define PHILOSOPHS 5		/*Pocet filosofu*/
#define MAX_LENGTH 30		/*Maximalni delka zprav*/
#define THINKING_TIME 5000000	/*Maximalni cas pri simulaci premysleni*/
#define EATING_TIME 5000000      /*Maximalni cas pri simulaci jidla*/
#ifndef TRUE
  #define TRUE 1
#endif
#ifndef FALSE
  #define FALSE 0
#endif
#define TH_TEXT "Think"		/*Oznameni o zmene stavu na THINKING*/
#define EA_TEXT "Eat"		/*Oznameni o zmene stavu na EATING*/
#define THINKING 0		
#define EATING 1
#define LEFT 1			/*Identifikace leveho souseda ve zprave*/
#define RIGHT 2			/*Identifikace praveho souseda ve zprave*/

struct msgbuffer {
  long type;
  char text[MAX_LENGTH];
};

int pokracovat;			/*Probiha cinnost procesu, dokud je true*/
int msg_id[PHILOSOPHS];		/*Fronty zprav vsech filosofu, slouzi pro */
				/*jejich prirazeni a na konci zruseni*/
int msg_msg_id; 		/*Fronta zprav pro vypis cinnost*/
int msg_left;			/*Fronta pro zpravy levemu sousedovi*/
int msg_right;			/*Fronta pro zpravy pravemu sousedovi*/
int msg_my;			/*Fronta pro zpravy od sousedu*/
int state_left;			/*Stav leveho souseda*/
int state_right;		/*Stav praveho souseda*/

/*Zpracovani signalu SIG_INT*/
void sighndl(int sig) {
  pokracovat = FALSE;
}

/*Zpracovani signalu SIG_USR1*/
void sigusrhndl(int sig) {
  pokracovat = FALSE;
}

/*Simulace premysleni filozofa*/
void mysleni(){
  int sl_time;
  sl_time = rand()%THINKING_TIME;
  usleep(sl_time);
}

/*Simulace jidla*/
void jidlo(){
  int sl_time;
  sl_time = rand()%EATING_TIME;
  usleep(sl_time);
}

/*Vytvoreni fronty zprav*/
int make_msgq(key_t key){
  int msgq_id;
  if((msgq_id = msgget(key, IPC_CREAT|0666)) == -1){
    fprintf(stderr,"Nelze vytvorit frontu zprav");
    exit(0);
  }
  return msgq_id;
}

/*Zruseni fronty zprav*/
void destroy_msgq(int msgq){
  msgctl(msgq,IPC_RMID,0);
}

/*Odeslani zpravy*/
void send_message(int msgqid, long num, char *text){
  struct msgbuffer buffer;
  buffer.type = num;
  strncpy(buffer.text,text,MAX_LENGTH);
  if(msgsnd(msgqid,(struct msgbuf *)&buffer,strlen(buffer.text)+1,0) == -1){
    fprintf(stderr,"Nepodarilo se odeslat zpravu");
    exit(-1);
  }
}

/*Vyzvednuti zpravy z fronty. Pokud neni zprava k dispozici, pocka na ni*/
int rcv_message(int msgqid, long num, char *text){
  int retval;
  struct msgbuffer buffer;
  buffer.type = num;
  retval = msgrcv(msgqid,(struct msgbuf *)&buffer,MAX_LENGTH,num,0);
  strncpy(text,buffer.text,MAX_LENGTH);
  if (retval != -1) {
    retval = buffer.type;
  }
  return retval;
}

/*Vyzvednuti zpravy z fronty. Pokud neni zprava k dispozici, vraci -1*/
int nw_rcv_message(int msgqid, long num, char *text){
  int retval;
  struct msgbuffer buffer;
  buffer.type = num;
  retval = msgrcv(msgqid,(struct msgbuf *)&buffer,MAX_LENGTH,num,IPC_NOWAIT);
  strncpy(text,buffer.text,MAX_LENGTH);
  if (retval != -1) {
    retval = buffer.type;
  }
  return retval;
}

/*Ceka, dokud nejsou pravy i levy soused ve stavu THINKING. Potom jim oznami
  zmenu stavu na EATING*/
void takeForks(void){
  char text[MAX_LENGTH];
  int typ;
 
  do {      /*Dokud oba sousede nejsou ve stavu THINKING*/
 	    /*Dokud jsou zpravy ve fronte*/
    while((typ = nw_rcv_message(msg_my, 0, text)) != -1){
      if(typ == LEFT){		/*Pokud zpravu poslal levy soused*/
        if(strcmp(text,EA_TEXT) ==0){
          state_left = EATING;		/*Zapamatuje si stav leveho souseda*/
        }
        if(strcmp(text,TH_TEXT) ==0){
          state_left = THINKING;	/*Zapamatuje si stav leveho souseda*/
        }
      }
      if(typ == RIGHT){		/*Pokud zpravu poslal levy soused*/
        if(strcmp(text,EA_TEXT) ==0){
          state_right = EATING;		/*Zapamatuje si stav praveho souseda*/
        }
        if(strcmp(text,TH_TEXT) ==0){
          state_right = THINKING;	/*Zapamatuje si stav praveho souseda*/
        }
      }
    }
  } while(state_left == EATING || state_right == EATING);

  /*Oznami levemu sousedovi zmenu stavu. Typ je RIGHT, protoze pro nej je tento
    pravym*/
  send_message(msg_left, RIGHT, EA_TEXT);
  /*Oznami pravemu sousedovi zmenu stavu. Typ je LEFT, protoze pro nej je tento
    levym*/
  send_message(msg_right, LEFT, EA_TEXT);
}

/*Oznami sousedum prechod do stavu THINKING*/
void putForks(void){
  send_message(msg_left, RIGHT, TH_TEXT);
  send_message(msg_right, LEFT, TH_TEXT);
}

/*Simulace jednoho filozofa*/
void philosoph(int whoami){
  char buffer[MAX_LENGTH];
  pokracovat == TRUE;
  send_message(msg_msg_id, whoami,"Narodil jsem se a myslim"); /*Vypis*/
  while(pokracovat == TRUE){	/*Promenna zmeni hodnotu v obsluze preruseni*/
    mysleni();			
    send_message(msg_msg_id, whoami,"Chci jist"); 	/*Vypis*/
    takeForks();	/*Pocka, az budou oba sousede ve stavu THINKING*/
    send_message(msg_msg_id, whoami,"Zacinam jist");	 /*Vypis*/
    jidlo();
    send_message(msg_msg_id, whoami,"Jdu premyslet");	 /*Vypis*/
    putForks();		/*Oznami sousedum zmenu stavu*/
  }
  send_message(msg_msg_id, whoami,"Umiram"); 	/*Vypis*/
}

/*Vypisuje cinnosti filozofu na standardni vystup*/
void logger(){
  int phil;
  char buffer[MAX_LENGTH];
  while (pokracovat == TRUE) {/*Promenna zmeni hodnotu v obsluze preruseni*/
    if((phil=nw_rcv_message(msg_msg_id,0,buffer))!=-1){
      printf("%d: %s\n",phil,buffer);
    }
  }
}


/*Vypisuje cinnosti filozofu na standardni vystup. Dojdou-li zpravy ve fronte,
  funkce konci*/
void clearMsg(){
  int phil;
  char buffer[MAX_LENGTH];
  while((phil=nw_rcv_message(msg_msg_id,0,buffer))!=-1){
    printf("%d: %s\n",phil,buffer);
  }
}


int main(int argc, char **argv){
  int philids[PHILOSOPHS];	/*ID vsech filozofu pro jejich ukonceni*/
  int i,j,vysl;
  char znak1;
  key_t key;

  srand((unsigned) time(NULL));	/*Inicializace generatoru nahodnych cisel*/
  znak1=(char)(rand()%255);     /*Nahodny znak*/
  key = ftok(argv[0],znak1); /*Nahodny klic pro semafor*/
  msg_msg_id=make_msgq(key);	/*Vytvori frontu pro vypisy o cinnosti*/
  for(i=0;i<PHILOSOPHS;i++){
    znak1=(znak1+1)%255;	/*Kazda fronta ma jiny znak pro klic*/
    key = ftok(argv[0],znak1); /*Nahodny klic pro frontu*/
    msg_id[i]=make_msgq(key);  /*Vytvori fronty pro filozofy*/
    philids[i]=-1;		/*Dokud nejsou vytvoreni filozofove, je */
				/*jejich cislo procesu -1*/
  }
  pokracovat = TRUE;

  /*Vytvoreni synovskych procesu - filozofu*/
  for (i=0;i<PHILOSOPHS;i++){
    msg_left = msg_id[((i+PHILOSOPHS)-1)%PHILOSOPHS]; /*Fronta leveho souseda*/
    msg_right = msg_id[(i+1)%PHILOSOPHS];	/*Fronta praveho souseda*/
    msg_my = msg_id[i];				/*Fronta tohoto filozofa*/
    state_right=state_left=THINKING;		/*Stavy sousedu*/
    if((philids[i]=fork())==-1){
      fprintf(stderr,"Nelze vytvorit synovsky proces.\n");
      exit(0);
    } else if (philids[i] == 0) {		/*Synovsky proces - filozof*/
      srand((unsigned) time(NULL));	
      	 			/*Inicializace generatoru nahodnych cisel*/
      signal(SIGINT,sighndl);			/*Obsluha signalu SIG_INT*/
      signal(SIGUSR1,sigusrhndl);		/*Obsluha signalu SIG_INT*/
      philosoph(i+1);	/*Cinnost filozofa. Parametr s jeho ID slouzi pro */
			/*vypisy na obrazovku*/
      break;
    } else {		      /*Otcovsky proces*/
      if(i<PHILOSOPHS-1){	/*Jeste nebyli vytvoreni vsichni filozofove*/	
	continue;
      }
      else {			/*Jiz jsou vytvoreni vsichni filozofove*/
        signal(SIGINT,sighndl);		/*Obsluha signalu SIG_INT*/
        logger();			/*Vypisy na obrazovku do prichodu*/
					/*signalu SIG_INT*/
        for(j=0;j<PHILOSOPHS;j++){	/*Upozorni filozofy signalem*/
          if(philids[j]!=-1){
            kill(philids[j],SIGUSR1);
          }
        }
        for(j=0;j<PHILOSOPHS;j++){	/*Pocka na konec vsech filozofu*/
          if(philids[j]!=-1){
            waitpid(philids[j],&vysl,0);
          }
        }
        clearMsg();			/*Vypise zbytek informaci od filozofu*/
      }
      for(i=0;i<PHILOSOPHS;i++){	/*Zrusi fronty zprav filozofu*/
        destroy_msgq(msg_id[i]);
      }
      destroy_msgq(msg_msg_id);	/*Zrusi frontu zprav na vypisy*/
      break;
    }
  }
  return 0;
}
    

