/******************************************************************************

 Autor: Martin Kot
 Studijni cislo: KOT119
 Predmet: Operacni systemy

 Projekt cislo 3 - daemon pro spousteni programu v urceny cas
 
 Program nacte soubor s casovymi udaji a k nim prirazenymi prikazy. Zaznamy 
 s casem, ktery jiz byl, vykona hned a prejde do rezimu daemona. V nem spousti
 dalsi prikazy v uvedenych casech a jejich vystup presmerovava do souboru. Po 
 vykonani vsech program konci.

******************************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <signal.h>
#include "proj3.h"


FILE *input;       //Vstupni soubor
char *yytext=NULL; //Buffer pro cteni vstupniho souboru
int yylineno = 1;  //Cislo aktualne cteneho radku souboru
char *filename;    //Jmeno vstupniho souboru
int length=0;      //Delka retezce v bufferu
zaznam *data;      //Pole zaznamu s casy a prikazy
int buffer_len;    //Delka bufferu
int pokracovat;    //Slouzi pro zpracovani signalu ALARM - nastavi ji na 0

/*Vytiskne oznameni o chybe ve vstupnim souboru*/
void pError(char *msg){
   fprintf(stderr,"%s:%d: %s\n",filename,yylineno,msg);
   exit(-1);
}

/*Zpracovani signalu*/
void sighndl(int cisloSig){
  pokracovat=FALSE;
}

/*Prochazi vstupni soubor a hleda v nem lexikalni jednotky*/
int lex(int prikaz){
  int i=0;
  int c;
  if(length==0 || yytext ==NULL){
    if((yytext=(char *)malloc(BUF_LEN*sizeof(char))) == NULL){
      fprintf(stderr,"Nedostatek pameti pro beh programu.");
      exit(-1);
    }
    else{
      length=BUF_LEN*sizeof(char);
    }
  }
  if(!prikaz) {
    while(TRUE){
      c=fgetc(input);
      if(c <= '9' && c>='0'){
        yytext[i]=c;
        i++;
        c=fgetc(input);
        while(c<='9' && c>='0'){
          yytext[i]=c;
          i++;
          if(length==i){
            if((void*)realloc(yytext,length+(BUF_LEN*sizeof(char)))==NULL){
              fprintf(stderr,"Nedostatek pameti pro beh programu.");
              exit(-1);
            }
            else{
              length+=BUF_LEN*sizeof(char);
            }
          }
          c=fgetc(input);
        }
        ungetc(c,input);
        yytext[i]='\0';
        return CISLO;
      }
      else if(c==':'){
        return DVOJTECKA;
      }
      else if(c==' ' ||c == '\r') {
        continue;
      }
      else if(c == '\n') {
        yylineno++;
        continue;
      }
      else if(c == EOF) {
        return KONEC;
      }
      else {
        fprintf(stderr,"%s:%d Neznamy znak %c\n",filename,yylineno,c);
        exit(-1);
      }
    }
  }
  else{
    while(TRUE){
      c=fgetc(input);
      if(c=='\n'){
        yylineno++;
        yytext[i]='\0';
        return PRIKAZ;
      }
      if(c == EOF){
        yytext[i]='\0';
        return PRIKAZ;
      }
      else {
        yytext[i]=c;
        i++;
        if(length==i){
          if((void*)realloc(yytext,length+(BUF_LEN*sizeof(char)))==NULL){
            fprintf(stderr,"Nedostatek pameti pro beh programu.");
            exit(-1);
          }
          else{
            length+=BUF_LEN*sizeof(char);
          }
        }
      }
    }
  }
}        
        
/*Vytvori pole zaznamu s casy a prikazy podle souboru*/
int parseFile(void){
  int i=0;
  int h,m;

  int actSymbol;
  while(TRUE){
    /*Hodina*/
    if((actSymbol=lex(FALSE))!=CISLO) {
        if(actSymbol==KONEC) break;
        else
	  pError("Ocekavana hodina");
    }
    if(strlen(yytext)>2) {
      pError("Hodnota hodiny musi byt dvojciferna");
    }
    h=atoi(yytext);
    data[i].hod=atoi(yytext);
    if(data[i].hod>23) {
      pError("Hodnota hodiny je prilis velka");
    }

    /*Dvojtecka*/
    if(lex(FALSE)!=DVOJTECKA) {
      pError("Ocekavana dvojtecka");
    }

    /*Minuta*/
    if(lex(FALSE)!=CISLO) {
      pError("Ocekavana minuta");
    }
    if(strlen(yytext)>2) {
      pError("Hodnota minuty musi byt dvojciferna");
    }
    data[i].min=atoi(yytext);
    if(data[i].min>59) {
      pError("Hodnota minuty je prilis velka");
    }

    /*Prikaz*/
    if(lex(TRUE)!=PRIKAZ) {
      pError("Ocekavan prikaz");
    }
    data[i].prikaz = yytext;
    yytext = NULL;
    i++;

    /*Prodlouzeni bufferu*/
    if(i>=buffer_len){
      buffer_len+=DATA_COUNT;    
      if((void *)realloc(data,buffer_len*sizeof(zaznam))==NULL){
        fprintf(stderr,"Nedostatek pameti\n");
        exit(-1);
      }
    }
  }
  return i;
}

/*Porovna dva zaznamy podle casu*/
int compare(void *a, void *b){
  int c1,c2;
  c1 = ((zaznam *)a)->hod * 60 + ((zaznam *)a)->min;
  c2 = ((zaznam *)b)->hod * 60 + ((zaznam *)b)->min;
  return c1-c2;
}

/*Vykona prikaz zadany jako retezec*/
int provedPrikaz(char *prik){
  pid_t pid;
  int retval;

  if((pid=fork())==-1){
    fprintf(stderr,"Nelze vytvorit synovsky proces");
  } else if (pid == 0) {
    execlp(SHELL,SHELL,"-c",prik,NULL); 
  } else {
    waitpid(pid,&retval,0);
  }
  return retval;
}

/*Porovna aktualni cas s casem v zaznamu*/
int staryCas(zaznam casPrik){
  struct tm *aktTime;
  time_t cas;
  time(&cas);
  aktTime=localtime(&cas);
  if(casPrik.hod < aktTime->tm_hour){
    return TRUE;
  }
  if(casPrik.hod > aktTime->tm_hour){
    return FALSE;
  }
  if(casPrik.min <= aktTime->tm_min){
    return TRUE;
  }
  return FALSE;
}

/*Pocet sekund chybejicich do zadaneho casu vykonani prikazu*/
int chybiSekund(zaznam casPrik){
  struct tm *aktTime;
  time_t cas;
  time(&cas);
  aktTime=localtime(&cas);
  return ((casPrik.hod*60+casPrik.min) - 
	(aktTime->tm_hour*60+aktTime->tm_min))*60 - aktTime->tm_sec;
}

/*Zapise zacatek vykonavani prikazu do log souboru*/
void logger(FILE *logfile, zaznam prikZaz){
  struct tm *aktTime;
  time_t cas;
  time(&cas);
  aktTime=localtime(&cas);
  fprintf(logfile,"%.2d:%.2d:%.2d Zacal: %s\n",aktTime->tm_hour,aktTime->tm_min,
	aktTime->tm_sec,prikZaz.prikaz);
}

/*Zapise konec vykonavani prikazu do log souboru*/
void loggerEnd(FILE *logfile, int retVal){
  struct tm *aktTime;
  time_t cas;
  time(&cas);
  aktTime=localtime(&cas);
  fprintf(logfile,"%.2d:%.2d:%.2d Skoncil s navratovou hodnotou: %d\n",
    aktTime->tm_hour,aktTime->tm_min,aktTime->tm_sec,retVal);
}

/*Vypise napovedu k spusteni programu*/
void printUsage(void){ 
  printf("\nSpusteni programu: proj3 [jmeno] [-ojmeno] [-ejmeno] [-ljmeno]\n");
  printf("\nVolby: jmeno   - jmeno je nazev souboru s casy a prikazy\n");
  printf("       -ojmeno - jmeno je nazev souboru pro standardni vystup z "
	"programu\n");
  printf("       -ejmeno - jmeno je nazev souboru pro chybovy vystup z "
        "programu\n");
  printf("       -ljmeno - jmeno je nazev log souboru\n\n");
  
}

int main(int argc, char **argv){
  int data_count=0;
  pid_t pid;
  int retVal;
  int i, sec;
  FILE *log;
  char *outFilename=DEF_OUT;  //Jmeno souboru pro presmerovani stdout
  char *errFilename=DEF_ERR;  //Jmeno souboru pro presmerovani stderr
  char *logFilename=DEF_LOG;  //Jmeno log souboru
  filename = DEF_FILENAME;  
 
  /*Rozparsovani prikazove radky*/
  for(i=1;i<argc;i++){
    if(argv[i][0]=='-'){
      switch(argv[i][1]){
        case 'o':
            if(argv[i][2]!='\0'){
              outFilename = (argv[i]+2);
            }
            else {
              printUsage();
              exit(-1);
            }
            break;
        case 'e':
            if(argv[i][2]!='\0'){
              errFilename = (argv[i]+2);
            }
            else {
              printUsage();
              exit(-1);
            }
            break;
        case 'l':
            if(argv[i][2]!='\0'){
              logFilename = (argv[i]+2);
            }
            else {
              printUsage();
              exit(-1);
            }
            break;
        case 'h':
            printUsage();
            exit(-1);
        default:
            printUsage();
            exit(-1);
      }
    }
    else {
      filename=argv[i];
    }
  }
  
  /*Otevreni vstupniho souboru*/
  if((input=fopen(filename,"r"))==NULL){
    fprintf(stderr,"Nepodarilo se otevrit soubor %s\n",filename);
    exit(-1);
  }

  /*Alokace pameti pro buffer*/
  buffer_len=DATA_COUNT;
  if((data = (void *)malloc(buffer_len*sizeof(zaznam)))==NULL){
    fprintf(stderr,"Nedostatek pameti\n");
    exit(-1);
  }
  data_count=parseFile(); 
  if(yytext != NULL){
    free(yytext);
  }
  qsort(data,data_count,sizeof(zaznam),compare); //setridi zaznamy podle casu

  /*Provede zaznamy s casem jiz uplynulym*/
  for(i=0;i<data_count;i++){
    if(staryCas(data[i])==TRUE){
      provedPrikaz(data[i].prikaz);
    }
    else {
      break;
    }
  }

  if(i!=data_count){ //Jeste nektere zaznamy nebyly vykonany
     
    /*Presmeruje stderr, stdout a otevre log soubor*/
    if(freopen(outFilename,"w",stdout)==NULL){
      fprintf(stderr,"Nepodarilo se presmerovat standardni vystup");
      exit(-1);
    }
    if(freopen(errFilename,"w",stderr)==NULL){
      fprintf(stderr,"Nepodarilo se presmerovat chybovy vystup");
      exit(-1);
    }
    if((log=fopen(logFilename,"w"))==NULL){
      fprintf(stderr,"Nepodarilo se presmerovat chybovy vystup");
      exit(-1);
    }

    /*Skonci program a vytvori kopii na pozadi*/
    if((pid=fork())==-1){
      fprintf(stderr,"Nelze vytvorit synovsky proces");
      exit(-1);
    } else if (pid == 0) {
      setpgrp();
    } else if (pid!=0){
      exit(0);
    }
    
    /*Vykonava prikazy zadane v souboru*/
    while(i!=data_count){
      sec=chybiSekund(data[i]);
      if(sec<=0){  //Okamzite vykonani (cas jiz uplynul)
        logger(log,data[i]);
        provedPrikaz(data[i].prikaz);
        loggerEnd(log,retVal);
      }
      else {  
        pokracovat=TRUE;
        signal(SIGALRM,sighndl);
        alarm(sec);
        while(pokracovat==TRUE); //Pocka, az zpracovani alarmu nastavi na FALSE
        logger(log,data[i]);
        retVal=provedPrikaz(data[i].prikaz); //Vykona prikaz
        loggerEnd(log,retVal);
      }  
      i++;
    }
    fclose(log);
  }
}
